#![allow(clippy::missing_panics_doc)]
use std::{io::BufReader, sync::Arc};

use crate::{
    identification::{test_format, IdentifiedPeptideSource, PeaksData, PeaksVersion},
    modification::SimpleModificationInner,
    molecular_formula,
};

#[test]
fn peaks_x() {
    assert_eq!(
        test_format::<PeaksData>(
            BufReader::new(DATA_X.as_bytes()),
            None,
            false,
            true,
            Some(PeaksVersion::X),
        )
        .unwrap(),
        19
    );
}

#[test]
fn peaks_x_patched() {
    assert_eq!(
        test_format::<PeaksData>(
            BufReader::new(DATA_X_PATCHED.as_bytes()),
            None,
            false,
            true,
            Some(PeaksVersion::XPatched),
        )
        .unwrap(),
        19
    );
}

#[test]
fn peaks_x_patched_sep() {
    assert_eq!(
        test_format::<PeaksData>(
            BufReader::new(DATA_X_PATCHED_SEP.as_bytes()),
            None,
            false,
            true,
            Some(PeaksVersion::XPatched),
        )
        .unwrap(),
        3
    );
}

#[test]
fn peaks_x_plus() {
    assert_eq!(
        test_format::<PeaksData>(
            BufReader::new(DATA_XPLUS.as_bytes()),
            None,
            false,
            true,
            Some(PeaksVersion::XPlus),
        )
        .unwrap(),
        19
    );
}

#[test]
fn peaks_11() {
    assert_eq!(
        test_format::<PeaksData>(
            BufReader::new(DATA_11.as_bytes()),
            None,
            false,
            true,
            Some(PeaksVersion::V11),
        )
        .unwrap(),
        19
    );
}

#[test]
fn peaks_11_features() {
    match test_format::<PeaksData>(
        BufReader::new(DATA_11_FEATURES.as_bytes()),
        None,
        true,
        false,
        Some(PeaksVersion::V11Features),
    ) {
        Ok(n) => assert_eq!(n, 19),
        Err(e) => {
            println!("{e}");
            panic!("Failed identified peptides test");
        }
    }
}

#[test]
fn peaks_11_all_candidates() {
    assert_eq!(
        test_format::<PeaksData>(
            BufReader::new(DATA_11_ALL_CANDIDATES.as_bytes()),
            None,
            false,
            true,
            Some(PeaksVersion::V11),
        )
        .unwrap(),
        19
    );
}

#[test]
fn peaks_11_custom_modification() {
    assert_eq!(
        test_format::<PeaksData>(
            BufReader::new(DATA_11_CUSTOM_MODIFICATION.as_bytes()),
            Some(&vec![(
                Some(0),
                "oxidation".to_string(),
                Arc::new(SimpleModificationInner::Formula(molecular_formula!(O 1))),
            )]),
            false,
            true,
            Some(PeaksVersion::V11)
        )
        .unwrap(),
        19
    );
}

#[test]
fn peaks_12() {
    assert_eq!(
        test_format::<PeaksData>(
            BufReader::new(DATA_12.as_bytes()),
            None,
            false,
            true,
            Some(PeaksVersion::V12)
        )
        .unwrap(),
        19
    );
}

#[test]
fn peaks_ab() {
    assert_eq!(
        test_format::<PeaksData>(
            BufReader::new(DATA_AB.as_bytes()),
            None,
            false,
            true,
            Some(PeaksVersion::Ab)
        )
        .unwrap(),
        19
    );
}

#[test]
fn peaks_x_plus_sep() {
    assert_eq!(
        test_format::<PeaksData>(
            BufReader::new(DATA_X_PLUS_SEP.as_bytes()),
            None,
            false,
            true,
            Some(PeaksVersion::XPlus)
        )
        .unwrap(),
        19
    );
}

#[test]
fn peaks_db_peptide() {
    assert_eq!(
        test_format::<PeaksData>(
            BufReader::new(DATA_DB_PEPTIDE.as_bytes()),
            None,
            false,
            false,
            Some(PeaksVersion::DBPeptide)
        )
        .unwrap(),
        19
    );
}

#[test]
fn peaks_db_psm() {
    assert_eq!(
        test_format::<PeaksData>(
            BufReader::new(DATA_DB_PSM.as_bytes()),
            None,
            false,
            false,
            Some(PeaksVersion::DBPSM)
        )
        .unwrap(),
        19
    );
}

#[test]
fn peaks_db_protein_peptide() {
    assert_eq!(
        test_format::<PeaksData>(
            BufReader::new(DATA_DB_PROTEIN_PEPTIDE.as_bytes()),
            None,
            false,
            false,
            Some(PeaksVersion::DBProteinPeptide)
        )
        .unwrap(),
        19
    );
}

#[test]
fn full_peaks_file() {
    for pep in PeaksData::parse_file("data/200305_HER_test_04_DENOVO_excerpt.csv", None).unwrap() {
        if let Err(e) = pep {
            panic!("{}", e);
        }
    }
}

#[test]
fn fuzz_crashes() {
    let mut all_passing = true;
    for entry in std::fs::read_dir("src/identification/peaks_fuzz_tests").unwrap() {
        let entry = entry.unwrap();
        let path = entry.path();
        let metadata = std::fs::metadata(&path).unwrap();
        let name = path.file_name().unwrap().to_str().unwrap();
        if metadata.is_file() && name.starts_with("crash") {
            match std::panic::catch_unwind(|| {
                test_format::<PeaksData>(
                    BufReader::new(std::fs::File::open(&path).unwrap()),
                    None,
                    false,
                    true,
                    Some(PeaksVersion::V11),
                )
            }) {
                Ok(_) => {
                    println!("identification::peaks_tests::fuzz_crashes::{name} passed");
                }
                Err(err) => {
                    all_passing = false;
                    println!("---- identification::peaks_tests::fuzz_crashes::{name} crash report ----\n{err:#?}\n");
                }
            }
        }
    }
    assert!(all_passing, "Some fuzz tests did not pass");
}

#[test]
fn fuzz_hangs() {
    let mut all_passing = true;
    for entry in std::fs::read_dir("src/identification/peaks_fuzz_tests").unwrap() {
        let entry = entry.unwrap();
        let path = entry.path();
        let metadata = std::fs::metadata(&path).unwrap();
        let name = path.file_name().unwrap().to_str().unwrap();
        if metadata.is_file() && name.starts_with("hang") {
            match std::panic::catch_unwind(|| {
                test_format::<PeaksData>(
                    BufReader::new(std::fs::File::open(&path).unwrap()),
                    None,
                    false,
                    true,
                    Some(PeaksVersion::V11),
                )
            }) {
                Ok(_) => {
                    println!("identification::peaks_tests::fuzz_hangs::{name} passed");
                }
                Err(err) => {
                    all_passing = false;
                    println!("---- identification::peaks_tests::fuzz_hangs::{name} crash report ----\n{err:#?}\n");
                }
            }
        }
    }
    assert!(all_passing, "Some fuzz tests did not pass");
}

const DATA_AB: &str = r"Scan,Peptide,Tag Length,ALC (%),length,m/z,z,RT,Area,Mass,ppm,Accession,PTM,local confidence (%),tag (>=0%),mode
F2:10351,MHQN(+.98)WLWL,8,98,8,564.7653,2,23.75,5.73E7,1127.5222,-5.5,,Deamidation (NQ),96 98 98 99 99 100 99 99,MHQN(+.98)WLWL,CID
F3:3063,M(+15.99)PHNHHTE,8,98,8,509.7123,2,10.99,4.42E6,1017.4087,1.4,,Oxidation (M),98 98 98 98 99 100 100 96,M(+15.99)PHNHHTE,CID
F3:3534,M(+15.99)PHNHHTE,8,98,8,509.7128,2,11.79,1.82E6,1017.4087,2.4,,Oxidation (M),98 99 99 98 99 99 100 95,M(+15.99)PHNHHTE,CID
F3:2117,HNHHTE,6,97,6,387.6678,2,8.52,2.03E6,773.3205,0.8,,,97 96 98 99 99 96,HNHHTE,CID
F2:13745,STMHWV,6,97,6,380.6751,2,30.62,2.73E6,759.3374,-2.2,,,98 99 95 95 99 99,STMHWV,CID
F1:20191,LFLN(+.98)ESHLTHAF,12,97,12,715.3615,2,37.09,4.4E6,1428.7036,3.3,,Deamidation (NQ),85 97 99 98 98 94 98 99 99 99 100 99,LFLN(+.98)ESHLTHAF,CID
F3:14603,APNTFTCSVLHE,12,97,12,659.8101,2,38.73,9.59E7,1317.6023,2.5,constructed_protein_Heavy,,96 96 98 99 100 99 96 95 96 99 98 94,APNTFTCSVLHE,CID
F3:15521,FTLNLHPVEEE,11,97,11,664.3321,2,41.05,2.65E6,1326.6455,3.1,,,99 99 99 97 99 99 90 92 99 100 94,FTLNLHPVEEE,CID
F4:8195,SNNYATHYAENK(+72.06),12,97,12,742.3466,2,18.77,6.26E7,1482.6792,-0.3,,Carboxyethyl,83 94 98 99 99 99 98 99 99 100 97 99,SNNYATHYAENK(+72.06),CID
F2:13141,WFVDLEEVHTA,11,97,11,673.3289,2,29.59,1.25E7,1344.6350,6.2,,,92 99 99 98 88 95 99 99 100 99 98,WFVDLEEVHTA,CID
F3:2323,HN(+.98)HHTE,6,97,6,388.1600,2,9.21,4.17E4,774.3045,1.2,,Deamidation (NQ),97 97 99 99 99 91,HN(+.98)HHTE,CID
F4:7286,ANN(+.98)HATYYAENK(+72.06),12,96,12,734.8412,2,17.49,7.73E6,1467.6682,-0.2,,Deamidation (NQ); Carboxyethyl,82 94 99 99 100 99 99 99 99 100 95 96,ANN(+.98)HATYYAENK(+72.06),CID
F2:7215,HYLHEV,6,96,6,399.1997,2,18.41,8.67E6,796.3868,-2.5,,,95 95 97 97 99 99,HYLHEV,CID
F1:20356,YLDQTEQWQLY,11,96,11,743.8483,2,37.34,1.04E7,1485.6775,3.1,,,90 97 99 96 99 99 96 98 94 97 96,YLDQTEQWQLY,CID
F3:14820,APNTFTCSVLHE,12,96,12,659.8107,2,39.24,6.09E6,1317.6023,3.5,constructed_protein_Heavy,,92 89 98 99 100 99 99 98 97 98 93 94,APNTFTCSVLHE,CID
F4:9062,SDNYATHYAENK(+72.06),12,96,12,742.8393,2,20.43,5.92E7,1483.6631,0.6,,Carboxyethyl,84 95 98 99 99 99 99 99 99 100 89 97,SDNYATHYAENK(+72.06),CID
F4:6305,NHATYYAENK(+72.06),10,96,10,641.8087,2,16.22,2.99E6,1281.6040,-1.0,,Carboxyethyl,85 96 99 99 99 98 98 100 94 94,NHATYYAENK(+72.06),CID
F2:8831,VCAAVHGV,8,96,8,378.1942,2,21.04,3.21E6,754.3796,-7.5,,,84 92 98 99 99 100 99 99,VCAAVHGV,CID
F4:3135,TPVSEHQK(+72.06),8,96,8,499.2701,2,10.87,3.28E7,996.5292,-3.5,,Carboxyethyl,99 99 99 97 99 94 88 95,TPVSEHQK(+72.06),CID";

const DATA_X: &str = r"Scan,Peptide,Tag Length,ALC (%),length,m/z,z,RT,Area,Mass,ppm,PTM,local confidence (%),tag (>=0%),mode
F4:14106,TTPPVLDSDGSFFLYSK,17,99,17,625.3127,3,55.62,8.89E6,1872.9146,1.0,,100 100 100 99 100 100 100 100 100 100 100 100 100 100 100 100 100,TTPPVLDSDGSFFLYSK,HCD
F1:14431,TTPPVLDSDGSFFLYSK,17,99,17,625.3127,3,55.41,6.07E6,1872.9146,0.9,,100 100 100 99 100 100 100 100 100 100 100 100 100 100 100 100 100,TTPPVLDSDGSFFLYSK,HCD
F3:3112,LSKADYEKHK,10,99,10,406.8876,3,16.39,2.6E4,1217.6404,0.6,,100 100 100 100 100 100 100 100 100 100,LSKADYEKHK,HCD
F3:6062,VTYDYYNK,8,99,8,533.2494,2,25.97,3.22E6,1064.4814,2.6,,100 100 100 100 100 100 99 100,VTYDYYNK,HCD
F3:4191,FGQGTKVELKR,11,99,11,421.5793,3,20.24,6.03E6,1261.7141,1.5,,100 100 100 100 100 100 100 100 100 100 100,FGQGTKVELKR,HCD
F4:10798,FNWYVDGVEVHNAK,14,99,14,839.4061,2,44.07,3.24E6,1676.7947,1.8,,99 100 100 100 100 100 100 100 100 100 100 99 100 100,FNWYVDGVEVHNAK,HCD
F1:4768,VSNKALPAPLEK,12,99,12,633.8753,2,23.51,1.19E6,1265.7344,1.3,,99 100 100 100 100 100 100 100 100 100 100 100,VSNKALPAPLEK,HCD
F1:14407,TTPPVLDSDGSFFLYSK,17,99,17,937.4662,2,55.35,1.15E8,1872.9146,1.8,,100 100 100 100 100 100 100 100 100 99 99 99 100 100 100 100 100,TTPPVLDSDGSFFLYSK,HCD
F1:15464,LLLYSASFLYSGVPSR,16,99,16,886.9857,2,59.69,2.04E8,1771.9509,3.4,,97 100 100 100 100 100 100 100 100 100 100 100 100 100 100 100,LLLYSASFLYSGVPSR,HCD
F1:5855,FTLSADTSK,9,99,9,485.2494,2,26.74,3.62E7,968.4814,2.9,,100 100 100 100 99 100 100 100 100,FTLSADTSK,HCD
F4:14140,TTPPVLDSDGSFFLYSK,17,99,17,937.4669,2,55.70,2.38E8,1872.9146,2.5,,100 100 100 100 100 100 100 100 100 99 99 99 100 100 100 100 100,TTPPVLDSDGSFFLYSK,HCD
F1:5902,FTLSADTSK,9,99,9,485.2482,2,26.88,,968.4814,0.4,,100 100 100 100 99 100 100 100 100,FTLSADTSK,HCD
F1:14222,TTPPVLDSDGSFFLYSK,17,99,17,937.4658,2,54.63,2.06E7,1872.9146,1.3,,100 100 100 100 100 100 100 100 100 99 98 99 100 100 100 100 100,TTPPVLDSDGSFFLYSK,HCD
F1:14461,TTPPVLDSDGSFFLYSK,17,99,17,937.4678,2,55.47,2.06E7,1872.9146,3.5,,100 100 100 100 100 100 100 100 100 98 98 99 100 100 100 100 100,TTPPVLDSDGSFFLYSK,HCD
F1:15517,LLLYSASFLYSGVPSR,16,99,16,886.9855,2,59.83,2.04E8,1771.9509,3.1,,97 100 100 100 100 100 100 100 100 100 100 99 99 100 100 100,LLLYSASFLYSGVPSR,HCD
F1:7735,QAPGKGLEWVAR,12,99,12,437.9118,3,32.98,4.64E6,1310.7095,3.1,,98 100 100 100 100 100 100 100 100 100 100 100,QAPGKGLEWVAR,HCD
F4:14079,TTPPVLDSDGSFFLYSK,17,99,17,937.4672,2,55.56,2.38E8,1872.9146,2.9,,100 100 100 100 100 100 100 100 100 98 98 99 100 100 100 100 100,TTPPVLDSDGSFFLYSK,HCD
F1:14468,TTPPVLDSDGSFFLYSK,17,99,17,937.4670,2,55.50,2.06E7,1872.9146,2.6,,100 100 100 100 100 100 100 100 100 98 98 99 100 100 100 100 100,TTPPVLDSDGSFFLYSK,HCD
F3:3851,FYPREAK,7,99,7,455.7434,2,19.26,5.57E6,909.4708,1.7,,100 100 99 99 100 100 100,FYPREAK,HCD";

const DATA_X_PATCHED_SEP: &str = r"sep=;
Fraction;Source File;Feature;Peptide;Scan;Tag Length;ALC (%);length;m/z;z;RT;Area;Mass;ppm;PTM;local confidence (%);tag (>=0%);mode
1;20190517_F1_Ag5_3117030_SA_ETHCD_131-2a_Tryp01.raw;F1:5056;LLYLVSK;F1:6994;7;99;7;418.2689;2;39.59;1.47E6;834.5215;2.2;;100 100 100 100 100 100 100;LLYLVSK;ETHCD
2;20190517_F1_Ag5_3117030_SA_ETHCD_131-2a_Chym01.raw;-;MYYSKL;F2:6253;6;99;6;402.7029;2;35.37;;803.3887;3.2;;100 100 100 100 100 100;MYYSKL;ETHCD
4;20190517_F1_Ag5_3117030_SA_ETHCD_131-2a_LysN01.raw;F4:9634;KAAVFNHFLSDGVK;F4:6537;14;99;14;511.6133;3;36.67;2.16E6;1531.8147;2.2;;100 100 100 100 100 100 100 100 100 100 100 100 100 100;KAAVFNHFLSDGVK;ETHCD";

const DATA_X_PATCHED: &str = r"Fraction,Source File,Feature,Peptide,Scan,Tag Length,ALC (%),length,m/z,z,RT,Area,Mass,ppm,PTM,local confidence (%),tag (>=0%),mode
1,20190517_F1_Ag5_3117030_SA_ETHCD_131-2a_Tryp01.raw,F1:5056,LLYLVSK,F1:6994,7,99,7,418.2689,2,39.59,1.47E6,834.5215,2.2,,100 100 100 100 100 100 100,LLYLVSK,ETHCD
2,20190517_F1_Ag5_3117030_SA_ETHCD_131-2a_Chym01.raw,-,MYYSKL,F2:6253,6,99,6,402.7029,2,35.37,,803.3887,3.2,,100 100 100 100 100 100,MYYSKL,ETHCD
4,20190517_F1_Ag5_3117030_SA_ETHCD_131-2a_LysN01.raw,F4:9634,KAAVFNHFLSDGVK,F4:6537,14,99,14,511.6133,3,36.67,2.16E6,1531.8147,2.2,,100 100 100 100 100 100 100 100 100 100 100 100 100 100,KAAVFNHFLSDGVK,ETHCD
6,20190517_F1_Ag5_3117030_SA_ETHCD_131-2a_TL01.raw,F6:4316,LLSDHRGATYA,F6:4340,11,99,11,401.8766,3,24.21,7.2E5,1202.6042,3.2,,100 100 100 100 100 100 100 100 100 100 100,LLSDHRGATYA,ETHCD
2,20190517_F1_Ag5_3117030_SA_ETHCD_131-2a_Chym01.raw,F2:137,KLAKVF,F2:5266,6,99,6,353.2372,2,29.28,6.29E7,704.4584,2.1,,100 100 100 100 100 100,KLAKVF,ETHCD
6,20190517_F1_Ag5_3117030_SA_ETHCD_131-2a_TL01.raw,F6:8687,LLKYASQS,F6:4154,8,99,8,455.2574,2,23.10,5.71E5,908.4967,3.9,,100 100 100 100 100 100 100 100,LLKYASQS,ETHCD
2,20190517_F1_Ag5_3117030_SA_ETHCD_131-2a_Chym01.raw,F2:137,KLAKVF,F2:5206,6,99,6,353.2372,2,29.28,6.29E7,704.4584,2.1,,100 100 100 100 100 100,KLAKVF,ETHCD
1,20190517_F1_Ag5_3117030_SA_ETHCD_131-2a_Tryp01.raw,F1:412,FTLSRDDSK,F1:4190,9,99,9,356.8489,3,23.11,1.72E6,1067.5247,0.3,,100 100 100 100 100 100 100 100 100,FTLSRDDSK,ETHCD
2,20190517_F1_Ag5_3117030_SA_ETHCD_131-2a_Chym01.raw,F2:6787,TLVGLVNY,F2:10516,8,99,8,439.7537,2,60.46,1.26E7,877.4909,2.3,,100 100 100 100 100 100 100 100,TLVGLVNY,ETHCD
6,20190517_F1_Ag5_3117030_SA_ETHCD_131-2a_TL01.raw,F6:16600,LTGTSTVGVGRGVLGDQKN,F6:5492,19,99,19,620.3372,3,31.03,2.2E7,1857.9907,-0.5,,100 100 99 99 100 100 100 100 100 100 100 100 100 100 100 100 99 100 100,LTGTSTVGVGRGVLGDQKN,ETHCD
1,20190517_F1_Ag5_3117030_SA_ETHCD_131-2a_Tryp01.raw,F1:10850,SSTLTLTKDEYER,F1:5255,13,99,13,514.9270,3,29.45,5.87E5,1541.7573,1.2,,100 100 100 100 100 100 100 100 100 100 99 100 100,SSTLTLTKDEYER,ETHCD
6,20190517_F1_Ag5_3117030_SA_ETHCD_131-2a_TL01.raw,F6:12891,LRSSVHYSQGYNNA,F6:3949,14,99,14,532.5908,3,21.95,1.03E7,1594.7488,1.2,,99 99 100 100 100 100 100 100 100 100 100 100 100 100,LRSSVHYSQGYNNA,ETHCD
6,20190517_F1_Ag5_3117030_SA_ETHCD_131-2a_TL01.raw,F6:1122,LRVEKKNW(+15.99),F6:3772,8,99,8,363.5458,3,20.31,1.99E6,1087.6138,1.8,Oxidation (HW),100 99 100 100 100 100 100 100,LRVEKKNW(+15.99),ETHCD
5,20190517_F1_Ag5_3117030_SA_ETHCD_131-2a_Ela01.raw,F5:11029,PYTFGGGTKLELKRA,F5:6127,15,99,15,546.6385,3,34.73,4.04E6,1636.8936,0.2,,99 99 100 100 100 100 100 100 100 100 100 100 100 100 100,PYTFGGGTKLELKRA,ETHCD
6,20190517_F1_Ag5_3117030_SA_ETHCD_131-2a_TL01.raw,F6:12937,VTYDYYKGG,F6:4919,9,99,9,533.2510,2,27.58,1.28E7,1064.4814,5.6,,99 99 100 100 100 100 100 100 100,VTYDYYKGG,ETHCD
4,20190517_F1_Ag5_3117030_SA_ETHCD_131-2a_LysN01.raw,F4:12679,KASGFTFTDFSLHMK(+58.01),F4:11257,15,99,15,592.2878,3,62.97,4.75E6,1773.8396,1.2,Carboxymethyl (KW  X@N-term),99 100 100 100 100 100 100 100 100 100 100 100 100 100 100,KASGFTFTDFSLHMK(+58.01),ETHCD
6,20190517_F1_Ag5_3117030_SA_ETHCD_131-2a_TL01.raw,F6:15331,VFTEQADLSGLTETKK,F6:6459,16,99,16,589.6451,3,36.75,9.07E5,1765.9097,2.2,,100 100 100 100 99 100 100 100 100 100 100 100 100 100 100 100,VFTEQADLSGLTETKK,ETHCD
5,20190517_F1_Ag5_3117030_SA_ETHCD_131-2a_Ela01.raw,F5:6860,TFGAGTKLELKRA,F5:5414,13,99,13,464.6069,3,30.47,2.49E6,1390.7932,4.0,,100 100 100 100 100 100 100 100 100 100 100 100 100,TFGAGTKLELKRA,ETHCD
2,20190517_F1_Ag5_3117030_SA_ETHCD_131-2a_Chym01.raw,F2:7876,TLSRDDSKSSVY,F2:3962,12,99,12,453.2252,3,21.77,3.86E6,1356.6521,1.1,,100 100 100 100 100 100 100 100 100 99 100 100,TLSRDDSKSSVY,ETHCD";

const DATA_XPLUS: &str = r"Fraction,Source File,Feature,Peptide,Scan,Tag Length,Denovo Score,ALC (%),length,m/z,z,RT,Predict RT,Area,Mass,ppm,PTM,local confidence (%),tag (>=0%),mode
10,20191211_F1_Ag5_peng0013_SA_her_Asp_N.raw,F10:3434,DYEKHKVYAC(+58.01),F10:3629,10,99,99,10,438.5332,3,19.91,-,2.3176E6,1312.5757,1.6,Carboxymethyl,100 100 100 100 100 100 100 100 100 100,DYEKHKVYAC(+58.01),ETHCD
4,20191211_F1_Ag5_peng0013_SA_her_Ela.raw,F4:4797,SGFGGLKN(+.98)TYLHW,F4:9505,13,99,99,13,494.2459,3,52.43,-,2.4924E7,1479.7146,0.9,Deamidation (NQ),100 100 100 100 100 100 100 100 100 100 100 100 100,SGFGGLKN(+.98)TYLHW,HCD
3,20191211_F1_Ag5_peng0013_SA_her_thermo.raw,F3:12703,LSC(+58.01)AASGFNLKDTY,F3:7983,14,99,99,14,774.3562,2,43.80,-,7.2888E7,1546.6973,0.4,Carboxymethyl,99 100 100 100 100 100 100 100 100 100 100 100 100 100,LSC(+58.01)AASGFNLKDTY,HCD
11,20191211_F1_Ag5_peng0013_SA_her_CB.raw,F11:14212,DSTYSLSSTLTLSK,F11:8043,14,99,99,14,751.8829,2,44.29,-,3.8265E6,1501.7512,0.0,,99 99 100 100 100 100 100 100 100 100 100 100 100 100,DSTYSLSSTLTLSK,ETHCD
2,20191211_F1_Ag5_peng0013_SA_her_chymo.raw,F2:3799,SGFGGLKDTYLHW,F2:9542,13,99,99,13,494.2458,3,52.79,-,5.6351E5,1479.7146,0.6,,100 100 100 100 100 100 100 100 100 100 100 100 100,SGFGGLKDTYLHW,HCD
11,20191211_F1_Ag5_peng0013_SA_her_CB.raw,F11:7245,SSPVTKSFNRGEC(+58.01),F11:4201,13,99,99,13,490.5618,3,22.70,-,3.2875E6,1468.6616,1.3,Carboxymethyl,100 100 100 100 100 100 100 100 100 100 100 100 100,SSPVTKSFNRGEC(+58.01),ETHCD
3,20191211_F1_Ag5_peng0013_SA_her_thermo.raw,-,VGVGRGVLGDQKN,F3:4406,13,99,99,13,433.5779,3,24.14,-,0,1297.7102,1.2,,100 100 100 100 100 100 100 100 100 100 100 100 100,VGVGRGVLGDQKN,ETHCD
7,20191211_F1_Ag5_peng0013_SA_her_Arg_C.raw,F7:14232,YSSDEKVLGEDFSDTR,F7:6514,16,99,99,16,616.6145,3,35.10,-,7.3986E5,1846.8220,-0.2,,100 100 100 100 100 100 100 100 100 100 100 100 100 100 100 98,YSSDEKVLGEDFSDTR,HCD
2,20191211_F1_Ag5_peng0013_SA_her_chymo.raw,F2:4750,TLSKADYEKHKVY,F2:3789,13,99,99,13,527.9471,3,20.26,-,3.2229E9,1580.8198,-0.3,,99 100 100 100 100 100 100 100 100 100 100 100 100,TLSKADYEKHKVY,ETHCD
5,20191211_F1_Ag5_peng0013_SA_her_aLP.raw,F5:10958,TPPVLDSDGSFFLYSKLT,F5:11087,18,99,99,18,663.0070,3,62.15,-,6.1345E6,1985.9985,0.3,,98 99 100 100 100 100 100 100 100 100 100 100 100 100 100 100 100 100,TPPVLDSDGSFFLYSKLT,ETHCD
4,20191211_F1_Ag5_peng0013_SA_her_Ela.raw,F4:6762,S(+58.01)GFNLKDTYLHWV,F4:10815,13,99,99,13,546.6037,3,60.06,-,2.739E7,1636.7886,0.4,Carboxymethyl (KW  X@N-term),100 100 100 100 100 100 100 100 100 100 100 99 99,S(+58.01)GFNLKDTYLHWV,HCD
8,20191211_F1_Ag5_peng0013_SA_her_Lys_C.raw,F8:5426,FNWYVDGVEVHDAK,F8:8307,14,99,99,14,560.2675,3,46.34,-,3.56E8,1677.7786,1.3,,99 100 100 100 100 100 100 100 100 100 100 100 100 100,FNWYVDGVEVHDAK,HCD
2,20191211_F1_Ag5_peng0013_SA_her_chymo.raw,F2:4750,TLSKADYEKHKVY,F2:3722,13,99,99,13,527.9471,3,20.26,-,3.2229E9,1580.8198,-0.3,,99 100 100 100 100 100 100 100 100 100 100 100 100,TLSKADYEKHKVY,ETHCD
4,20191211_F1_Ag5_peng0013_SA_her_Ela.raw,F4:7864,STLTLSKADYEKHKV,F4:4433,15,99,99,15,573.9813,3,24.22,-,6.8416E6,1718.9202,1.1,,99 99 100 100 100 100 100 100 100 100 100 100 100 100 100,STLTLSKADYEKHKV,ETHCD
11,20191211_F1_Ag5_peng0013_SA_her_CB.raw,F11:3224,LASYLDK,F11:4979,7,99,99,7,405.2239,2,26.98,-,6.041E4,808.4330,0.3,,100 100 100 100 100 100 99,LASYLDK,ETHCD
11,20191211_F1_Ag5_peng0013_SA_her_CB.raw,F11:9144,FNWYVDGVEVHNAK,F11:7980,14,99,99,14,559.9395,3,44.01,-,9.8729E4,1676.7947,1.2,,100 100 100 100 100 100 100 100 100 100 100 99 100 100,FNWYVDGVEVHNAK,HCD
8,20191211_F1_Ag5_peng0013_SA_her_Lys_C.raw,F8:8931,S(+58.01)TSGGTAALGC(+58.01)LVK,F8:7638,14,99,99,14,690.8375,2,42.67,-,2.888E7,1379.6602,0.2,Carboxymethyl (KW  X@N-term); Carboxymethyl,99 99 100 100 100 100 100 100 100 100 100 100 100 100,S(+58.01)TSGGTAALGC(+58.01)LVK,HCD
3,20191211_F1_Ag5_peng0013_SA_her_thermo.raw,F3:3555,LTLSKADYEKHK,F3:3745,12,99,99,12,478.2650,3,20.28,-,1.1906E8,1431.7722,0.7,,100 100 100 100 100 100 100 99 99 100 100 100,LTLSKADYEKHK,HCD
11,20191211_F1_Ag5_peng0013_SA_her_CB.raw,F11:12673,EVQLVESGGGLVQPGGSLRAK,F11:8595,21,99,99,21,694.3840,3,47.38,-,7.0626E7,2080.1274,1.4,,98 98 99 100 100 100 100 100 100 100 100 100 100 100 100 100 100 100 100 100 100,EVQLVESGGGLVQPGGSLRAK,ETHCD";

const DATA_11: &str = r#""Source File","Scan","Peptide","Tag length","ALC (%)","Length","m/z","z","RT","Area","Mass","ppm","PTM","local confidence (%)","mode",tag(>=0.0%),"Feature Id"
"20191211_F1_Ag5_peng0013_SA_her_Arg_C.raw",6514,YSSDEKVLGEDFSDTR,16,99.9,16,616.61450,3,35.1011,190625.72,1846.8220,-0.2,"",100 100 100 100 100 100 100 100 100 100 100 100 100 100 100 98,HCD,YSSDEKVLGEDFSDTR,14223
"20191211_F1_Ag5_peng0013_SA_her_Arg_C.raw",6743,DLQMTQSPSSLSASVGDR,18,99.2,18,626.96686,3,36.1993,794748.8,1877.8789,-0.1,"",99 99 100 100 100 100 99 99 100 100 100 100 100 100 100 100 100 90,ETHCD,DLQMTQSPSSLSASVGDR,14572
"20191211_F1_Ag5_peng0013_SA_her_Arg_C.raw",5644,DLQM(+15.99)TQSPSSLSASVGDR,18,99.0,18,947.94415,2,30.1547,548587.2,1893.8738,0.0,"Oxidation (M)",97 96 100 100 100 100 100 100 100 100 100 100 100 100 100 95 97 97,HCD,DLQM(+15.99)TQSPSSLSASVGDR,22869
"20191211_F1_Ag5_peng0013_SA_her_Arg_C.raw",6616,EVNSQFFR,8,98.8,8,513.75372,2,35.6492,60834.707,1025.4930,-0.1,"",98 97 99 100 99 100 100 97,HCD,EVNSQFFR,10540
"20191211_F1_Ag5_peng0013_SA_her_Arg_C.raw",9428,TYTFDNGTFLLR,12,98.7,12,724.36444,2,51.1627,83370.51,1446.7144,0.0,"",99 99 100 100 100 99 97 98 99 99 99 96,HCD,TYTFDNGTFLLR,17359
"20191211_F1_Ag5_peng0013_SA_her_Arg_C.raw",7805,EVQLVESGGGLVQPGGSLR,19,98.6,19,628.00598,3,42.2104,298160.3,1880.9955,0.3,"",98 97 99 100 100 100 100 100 100 100 100 100 98 99 99 99 100 100 83,ETHCD,EVQLVESGGGLVQPGGSLR,14598
"20191211_F1_Ag5_peng0013_SA_her_Arg_C.raw",5945,VLGEDFSDTR,10,98.0,10,569.77283,2,31.8568,18040.87,1137.5302,0.8,"",100 100 99 99 99 96 95 98 98 95,HCD,VLGEDFSDTR,12617
"20191211_F1_Ag5_peng0013_SA_her_Arg_C.raw",9483,TYTFDDGTFLLR,12,98.0,12,724.85663,2,51.3811,43949.74,1447.6984,0.3,"",96 97 100 100 100 100 96 96 97 99 99 95,HCD,TYTFDDGTFLLR,17378
"20191211_F1_Ag5_peng0013_SA_her_Arg_C.raw",3737,SRSGGGGNGLGSGGSLR,17,97.9,17,492.58188,3,19.8131,29103.092,1474.7236,0.1,"",100 99 100 100 100 99 98 95 100 100 100 100 98 98 99 99 83,ETHCD,SRSGGGGNGLGSGGSLR,9583
"20191211_F1_Ag5_peng0013_SA_her_Arg_C.raw",4286,TSGVLPR,7,97.8,7,365.21625,2,22.6110,744909.94,728.4181,-0.2,"",100 100 100 100 100 100 85,ETHCD,TSGVLPR,1074
"20191211_F1_Ag5_peng0013_SA_her_Arg_C.raw",6256,QAPGKGLEWVAR,12,97.8,12,437.91058,3,33.6266,,1310.7095,0.3,"",94 98 99 100 100 100 100 100 100 100 100 82,ETHCD,QAPGKGLEWVAR,0
"20191211_F1_Ag5_peng0013_SA_her_Arg_C.raw",4163,HEYTHYLQGR,10,97.6,10,435.21124,3,22.1258,218619.0,1302.6105,1.1,"",100 100 100 100 100 100 100 100 100 78,ETHCD,HEYTHYLQGR,6395
"20191211_F1_Ag5_peng0013_SA_her_Arg_C.raw",3647,TKESLSHFR,9,97.6,9,368.86447,3,19.3037,9257.09,1103.5724,-0.7,"",94 96 100 100 100 100 100 100 89,ETHCD,TKESLSHFR,1361
"20191211_F1_Ag5_peng0013_SA_her_Arg_C.raw",11907,LTWFDEGTAEFFAGSTR,17,97.6,17,967.94916,2,64.9324,66995.53,1933.8846,-0.4,"",87 87 100 100 100 100 100 100 100 100 100 100 99 99 97 96 95,HCD,LTWFDEGTAEFFAGSTR,23416
"20191211_F1_Ag5_peng0013_SA_her_Arg_C.raw",4773,DTLM(+15.99)LSR,7,97.6,7,426.21838,2,25.4288,115568.0,850.4218,0.4,"Oxidation (M)",95 95 100 100 100 100 94,HCD,DTLM(+15.99)LSR,5808
"20191211_F1_Ag5_peng0013_SA_her_Arg_C.raw",5396,GRGDSVVYGLR,11,97.6,11,393.54752,3,28.9093,54173.0,1177.6204,0.3,"",96 98 99 100 100 100 100 100 100 100 80,ETHCD,GRGDSVVYGLR,3284
"20191211_F1_Ag5_peng0013_SA_her_Arg_C.raw",8485,DSTYSLSSTLTSRR,14,97.5,14,787.40173,2,45.8671,242945.9,1572.7744,9.3,"",100 100 100 100 100 100 100 100 99 100 100 100 100 67,ETHCD,DSTYSLSSTLTSRR,19182
"20191211_F1_Ag5_peng0013_SA_her_Arg_C.raw",3880,ADSVFKGR,8,97.5,8,440.23810,2,20.6071,27530.41,878.4610,0.7,"",96 99 100 100 99 98 95 93,HCD,ADSVFKGR,6714
"20191211_F1_Ag5_peng0013_SA_her_Arg_C.raw",4217,HEYTHYLQAR,10,97.5,10,439.88297,3,22.4412,1841098.6,1316.6262,0.7,"",94 97 99 100 100 100 100 100 100 84,ETHCD,HEYTHYLQAR,6678"#;

const DATA_11_FEATURES: &str = r#""Source File","Feature Id","m/z","z","RT","RT Begin","RT End","Area","Quality","Denovo Peptide","ALC (%)","ppm"
"20191211_F1_Ag5_peng0013_SA_her_Arg_C.raw",14223,616.61450,3,35.1011,35.0728,35.1775,190625.72,0.00,YSSDEKVLGEDFSDTR,99.9,-0.2
"20191211_F1_Ag5_peng0013_SA_her_Arg_C.raw",14572,626.96686,3,36.1993,35.9724,36.6269,794748.8,0.00,DLQMTQSPSSLSASVGDR;DLQMTQSPSSLSASVGDR,99.2,-0.1
"20191211_F1_Ag5_peng0013_SA_her_Arg_C.raw",22869,947.94415,2,30.1547,29.8533,30.5004,548587.2,0.00,LDQM(+15.99)TQSPSSLSASVGDR;DLQM(+15.99)TQSPSSLSASVGDR,99.0,0.0
"20191211_F1_Ag5_peng0013_SA_her_Arg_C.raw",10540,513.75372,2,35.6492,35.6235,35.8589,60834.707,0.00,EVNSQFFR,98.8,-0.1
"20191211_F1_Ag5_peng0013_SA_her_Arg_C.raw",17359,724.36444,2,51.1627,51.0134,51.3298,83370.51,0.00,TYTFDNGTFLLR,98.7,0.0
"20191211_F1_Ag5_peng0013_SA_her_Arg_C.raw",14598,628.00598,3,42.2104,42.0118,42.5021,298160.3,0.00,EVQLVESGGGLVQPGGSLR,98.6,0.3
"20191211_F1_Ag5_peng0013_SA_her_Arg_C.raw",12617,569.77283,2,31.8568,31.7728,32.0192,18040.87,0.00,VLGEDFSDTR,98.0,0.8
"20191211_F1_Ag5_peng0013_SA_her_Arg_C.raw",17378,724.85663,2,51.3811,51.2976,51.5511,43949.74,0.00,TYTFDDGTFLLR,98.0,0.3
"20191211_F1_Ag5_peng0013_SA_her_Arg_C.raw",9583,492.58188,3,19.8131,19.7241,19.9643,29103.092,0.00,SRSGGGGNGLGSGGSLR,97.9,0.1
"20191211_F1_Ag5_peng0013_SA_her_Arg_C.raw",1074,365.21625,2,22.6110,22.4246,22.9395,744909.94,0.00,TSGVLPR;TSGVLPR,97.8,-0.2
"20191211_F1_Ag5_peng0013_SA_her_Arg_C.raw",6395,435.21124,3,22.1258,22.0404,22.4412,218619.0,0.00,HEYTHYLQGR,97.6,1.1
"20191211_F1_Ag5_peng0013_SA_her_Arg_C.raw",1361,368.86447,3,19.3037,19.2492,19.3908,9257.09,0.00,TKESLSHFR,97.6,-0.7
"20191211_F1_Ag5_peng0013_SA_her_Arg_C.raw",23416,967.94916,2,64.9324,64.7688,65.3386,66995.53,0.00,LTWFDEGTAEFFAGSTR,97.6,-0.4
"20191211_F1_Ag5_peng0013_SA_her_Arg_C.raw",5808,426.21838,2,25.4288,25.2470,25.7630,115568.0,0.00,DTLM(+15.99)LSR,97.6,0.4
"20191211_F1_Ag5_peng0013_SA_her_Arg_C.raw",3284,393.54752,3,28.9093,28.7614,29.1429,54173.0,0.00,GRGDSVVYGLR,97.6,0.3
"20191211_F1_Ag5_peng0013_SA_her_Arg_C.raw",19182,787.40173,2,45.8671,45.5640,46.2526,242945.9,0.00,DSTYSLSSGRFVVR;DSTYSLSSTLTSRR,97.5,9.3
"20191211_F1_Ag5_peng0013_SA_her_Arg_C.raw",6714,440.23810,2,20.6071,20.5252,20.8115,27530.41,0.00,ADSVFKGR,97.5,0.7
"20191211_F1_Ag5_peng0013_SA_her_Arg_C.raw",6678,439.88297,3,22.4412,22.3549,22.7358,1841098.6,0.00,HEYTHYLQAR;HEYTHYLQAR,97.5,0.7
"20191211_F1_Ag5_peng0013_SA_her_Arg_C.raw",16579,696.69293,3,42.1785,42.0943,42.3913,145613.02,0.00,VVGNDKALEVGNADDVLGFR,97.4,-3.7"#;

const DATA_11_ALL_CANDIDATES: &str = r#""Source File","Scan","Peptide","Tag length","ALC (%)","Length","m/z","z","RT","Area","Mass","ppm","PTM","local confidence (%)","mode",tag(>=0.0%),"Feature Id"
"20240903_SA_HTA_KR40_500ng_CID_130min.wiff",16281,Q(-17.03)PEDF,5,99.9,5,618.24072,1,48.7702,59523.16,617.2333,0.3,"Pyro-glu from Q",100 100 100 100 100,CID,Q(-17.03)PEDF,16481
"20240903_SA_HTA_KR40_500ng_CID_130min.wiff",16265,Q(-17.03)PEDF,5,99.9,5,618.24072,1,48.7702,59523.16,617.2333,0.3,"Pyro-glu from Q",100 100 100 100 100,CID,Q(-17.03)PEDF,16481
"20240903_SA_HTA_KR40_500ng_CID_130min.wiff",20270,LLGGPSVF,8,99.7,8,789.44788,1,72.3085,5838.5063,788.4432,-3.3,"",100 100 100 99 99 100 100 100,CID,LLGGPSVF,33201
"20240903_SA_HTA_KR40_500ng_CID_130min.wiff",9109,LVKDY,5,99.7,5,637.35510,1,18.7355,4015.6555,636.3483,-0.7,"",100 100 100 100 100,CID,LVKDY,18803
"20240903_SA_HTA_VE20_500ng_CID_60min.wiff",15195,DFATYYCQQHYTTPPTFGQGTKVE,24,99.7,24,1391.62524,2,33.1102,34988.38,2781.2380,-0.8,"",100 98 100 100 100 100 100 99 99 100 100 100 100 100 100 100 100 100 99 100 100 100 100 100,CID,DFATYYCQQHYTTPPTFGQGTKVE,29750
"20240903_SA_HTA_KR40_500ng_CID_90min.wiff",22518,LLGGPSVF,8,99.7,8,789.44580,1,71.3410,6788.68,788.4432,-5.9,"",100 100 100 99 99 100 100 100,CID,LLGGPSVF,32106
"20240903_SA_HTA_VE40_500ng_CID_60min.wiff",15635,APGKGLEWVARLYPTNGY,18,99.6,18,996.52295,2,34.7298,2675.0488,1991.0265,2.5,"",100 99 100 100 100 100 100 100 100 100 99 100 100 99 98 99 100 99,CID,APGKGLEWVARLYPTNGY,25167
"20240903_SA_HTA_KR40_500ng_CID_60min.wiff",11103,PEVKF,5,99.6,5,619.34204,1,23.4924,14385.308,618.3377,-4.7,"",99 100 100 100 99,CID,PEVKF,9777
"20240903_SA_HTA_KR20_500ng_CID_90min.wiff",15449,LLSDF,5,99.6,5,594.31116,1,50.1211,1865.4401,593.3060,-3.7,"",100 99 100 100 99,CID,LLSDF,8217
"20240903_SA_HTA_KR20_500ng_CID_90min.wiff",10862,VDVSHE,6,99.6,6,685.31506,1,25.3097,16523.75,684.3079,-0.1,"",100 100 100 100 98 100,CID,VDVSHE,16688
"20240903_SA_HTA_KR20_500ng_CID_60min.wiff",11421,DDYAGAMHWVR,11,99.5,11,660.79303,2,27.7775,31207.42,1319.5717,-0.1,"",99 98 99 100 100 100 100 100 100 100 100,CID,DDYAGAMHWVR,10245
"20240903_SA_HTA_VE40_500ng_CID_130min.wiff",16631,LLSDF,5,99.5,5,594.31287,1,51.6116,3949.231,593.3060,-0.8,"",100 99 100 100 99,CID,LLSDF,13293
"20240903_SA_HTA_KR40_500ng_CID_90min.wiff",16720,FLYSKL,6,99.5,6,770.44299,1,46.7960,14885.409,769.4374,-2.2,"",99 100 100 100 99 99,CID,FLYSKL,30214
"20240903_SA_HTA_KR40_500ng_CID_60min.wiff",13366,GYGVHWVR,8,99.5,8,487.25134,2,27.3181,,972.4930,-5.0,"",99 99 100 100 99 100 100 100,CID,GYGVHWVR,0
"20240903_SA_HTA_KR20_500ng_CID_130min.wiff",16083,GYGVHWVR,8,99.4,8,487.25189,2,45.3716,,972.4930,-3.8,"",99 99 100 99 99 100 100 100,CID,GYGVHWVR,0
"20240903_SA_HTA_KR40_500ng_CID_90min.wiff",16846,TFSGF,5,99.4,5,558.25446,1,47.0806,6637.823,557.2485,-2.5,"",98 99 100 100 100,CID,TFSGF,8188
"20240903_SA_HTA_KR40_500ng_CID_130min.wiff",15382,GYGVHWVR,8,99.3,8,487.25345,2,44.8251,,972.4930,-0.6,"",98 99 100 99 99 100 100 100,CID,GYGVHWVR,0
"20240903_SA_HTA_KR40_500ng_CID_60min.wiff",9567,VVDVSHED,8,99.3,8,899.41119,1,20.0530,831.1227,898.4032,0.8,"",100 99 100 100 100 98 99 99,CID,VVDVSHED,26075
"20240903_SA_HTA_VE20_500ng_CID_60min.wiff",15912,TKNQVSLTCLVKGF,14,99.3,14,1537.84521,1,34.7840,1830.2261,1536.8334,3.0,"",99 98 97 98 100 100 100 100 100 100 100 100 100 100,CID,TKNQVSLTCLVKGF,29989"#;

const DATA_11_CUSTOM_MODIFICATION: &str = r#""Source File","Scan","Peptide","Tag length","ALC (%)","Length","m/z","z","RT","Area","Mass","ppm","PTM","local confidence (%)","mode",tag(>=0.0%),"Feature Id"
"20191211_F1_Ag5_peng0013_SA_her_Arg_C.raw",6514,YSSDEKVLGEDFSDTR,16,99.9,16,616.61450,3,35.1011,190625.72,1846.8220,-0.2,"",100 100 100 100 100 100 100 100 100 100 100 100 100 100 100 98,HCD,YSSDEKVLGEDFSDTR,14223
"20191211_F1_Ag5_peng0013_SA_her_Arg_C.raw",6743,DLQMTQSPSSLSASVGDR,18,99.2,18,626.96686,3,36.1993,794748.8,1877.8789,-0.1,"",99 99 100 100 100 100 99 99 100 100 100 100 100 100 100 100 100 90,ETHCD,DLQMTQSPSSLSASVGDR,14572
"20191211_F1_Ag5_peng0013_SA_her_Arg_C.raw",5644,DLQM[C:Oxidation]TQSPSSLSASVGDR,18,99.0,18,947.94415,2,30.1547,548587.2,1893.8738,0.0,"Oxidation (M)",97 96 100 100 100 100 100 100 100 100 100 100 100 100 100 95 97 97,HCD,DLQM(+15.99)TQSPSSLSASVGDR,22869
"20191211_F1_Ag5_peng0013_SA_her_Arg_C.raw",6616,EVNSQFFR,8,98.8,8,513.75372,2,35.6492,60834.707,1025.4930,-0.1,"",98 97 99 100 99 100 100 97,HCD,EVNSQFFR,10540
"20191211_F1_Ag5_peng0013_SA_her_Arg_C.raw",9428,TYTFDNGTFLLR,12,98.7,12,724.36444,2,51.1627,83370.51,1446.7144,0.0,"",99 99 100 100 100 99 97 98 99 99 99 96,HCD,TYTFDNGTFLLR,17359
"20191211_F1_Ag5_peng0013_SA_her_Arg_C.raw",7805,EVQLVESGGGLVQPGGSLR,19,98.6,19,628.00598,3,42.2104,298160.3,1880.9955,0.3,"",98 97 99 100 100 100 100 100 100 100 100 100 98 99 99 99 100 100 83,ETHCD,EVQLVESGGGLVQPGGSLR,14598
"20191211_F1_Ag5_peng0013_SA_her_Arg_C.raw",5945,VLGEDFSDTR,10,98.0,10,569.77283,2,31.8568,18040.87,1137.5302,0.8,"",100 100 99 99 99 96 95 98 98 95,HCD,VLGEDFSDTR,12617
"20191211_F1_Ag5_peng0013_SA_her_Arg_C.raw",9483,TYTFDDGTFLLR,12,98.0,12,724.85663,2,51.3811,43949.74,1447.6984,0.3,"",96 97 100 100 100 100 96 96 97 99 99 95,HCD,TYTFDDGTFLLR,17378
"20191211_F1_Ag5_peng0013_SA_her_Arg_C.raw",3737,SRSGGGGNGLGSGGSLR,17,97.9,17,492.58188,3,19.8131,29103.092,1474.7236,0.1,"",100 99 100 100 100 99 98 95 100 100 100 100 98 98 99 99 83,ETHCD,SRSGGGGNGLGSGGSLR,9583
"20191211_F1_Ag5_peng0013_SA_her_Arg_C.raw",4286,TSGVLPR,7,97.8,7,365.21625,2,22.6110,744909.94,728.4181,-0.2,"",100 100 100 100 100 100 85,ETHCD,TSGVLPR,1074
"20191211_F1_Ag5_peng0013_SA_her_Arg_C.raw",6256,QAPGKGLEWVAR,12,97.8,12,437.91058,3,33.6266,,1310.7095,0.3,"",94 98 99 100 100 100 100 100 100 100 100 82,ETHCD,QAPGKGLEWVAR,0
"20191211_F1_Ag5_peng0013_SA_her_Arg_C.raw",4163,HEYTHYLQGR,10,97.6,10,435.21124,3,22.1258,218619.0,1302.6105,1.1,"",100 100 100 100 100 100 100 100 100 78,ETHCD,HEYTHYLQGR,6395
"20191211_F1_Ag5_peng0013_SA_her_Arg_C.raw",3647,TKESLSHFR,9,97.6,9,368.86447,3,19.3037,9257.09,1103.5724,-0.7,"",94 96 100 100 100 100 100 100 89,ETHCD,TKESLSHFR,1361
"20191211_F1_Ag5_peng0013_SA_her_Arg_C.raw",11907,LTWFDEGTAEFFAGSTR,17,97.6,17,967.94916,2,64.9324,66995.53,1933.8846,-0.4,"",87 87 100 100 100 100 100 100 100 100 100 100 99 99 97 96 95,HCD,LTWFDEGTAEFFAGSTR,23416
"20191211_F1_Ag5_peng0013_SA_her_Arg_C.raw",4773,DTLM[C:Oxidation]LSR,7,97.6,7,426.21838,2,25.4288,115568.0,850.4218,0.4,"Oxidation (M)",95 95 100 100 100 100 94,HCD,DTLM(+15.99)LSR,5808
"20191211_F1_Ag5_peng0013_SA_her_Arg_C.raw",5396,GRGDSVVYGLR,11,97.6,11,393.54752,3,28.9093,54173.0,1177.6204,0.3,"",96 98 99 100 100 100 100 100 100 100 80,ETHCD,GRGDSVVYGLR,3284
"20191211_F1_Ag5_peng0013_SA_her_Arg_C.raw",8485,DSTYSLSSTLTSRR,14,97.5,14,787.40173,2,45.8671,242945.9,1572.7744,9.3,"",100 100 100 100 100 100 100 100 99 100 100 100 100 67,ETHCD,DSTYSLSSTLTSRR,19182
"20191211_F1_Ag5_peng0013_SA_her_Arg_C.raw",3880,ADSVFKGR,8,97.5,8,440.23810,2,20.6071,27530.41,878.4610,0.7,"",96 99 100 100 99 98 95 93,HCD,ADSVFKGR,6714
"20191211_F1_Ag5_peng0013_SA_her_Arg_C.raw",4217,HEYTHYLQAR,10,97.5,10,439.88297,3,22.4412,1841098.6,1316.6262,0.7,"",94 97 99 100 100 100 100 100 100 84,ETHCD,HEYTHYLQAR,6678"#;

const DATA_12: &str = r#""Source File","Scan","Peptide","Tag length","Deep Novo Score (%)","ALC (%)","Length","m/z","z","RT","Area","Mass","ppm","PTM","local confidence (%)","mode",tag(>=0%),"Feature Id"
"9661_weiwei_mabs_novo_chymo.raw",4245;4251,TLSKADY,7,100.0,100.0,7,399.20514,2,20.9525,4.8573476E7,796.3967,-1.2,"",100 100 100 100 100 100 100,MIX,TLSKADY,1008
"9661_weiwei_mabs_novo_chymo.raw",4478;4483,SSPVTKSF,8,100.0,100.0,8,426.72626,2,27.9639,2.40281728E8,851.4388,-1.1,"",100 100 100 100 100 100 100 100,MIX,SSPVTKSF,1492
"9661_weiwei_mabs_novo_chymo.raw",3553;3556,TLSKADY,7,100.0,100.0,7,399.20514,2,20.9525,4.8573476E7,796.3967,-1.2,"",100 100 100 100 100 100 100,MIX,TLSKADY,1008
"9661_weiwei_mabs_novo_chymo.raw",5170;5177,TFRNY,5,100.0,100.0,5,350.67392,2,30.5363,8.4915336E7,699.3340,-1.0,"",100 100 100 100 100,MIX,TFRNY,5
"9661_weiwei_mabs_novo_chymo.raw",5046;5051,TFRNY,5,100.0,100.0,5,350.67392,2,30.5363,8.4915336E7,699.3340,-1.0,"",100 100 100 100 100,MIX,TFRNY,5
"9661_weiwei_mabs_novo_chymo.raw",6243;6249,SC(+58.01)SVM(+15.99)HEALH,10,100.0,100.0,10,396.49805,3,33.4147,4701863.5,1186.4747,-2.0,"Carboxymethyl; Oxidation (M)",100 100 100 100 100 100 100 100 100 100,MIX,SC(+58.01)SVM(+15.99)HEALH,931
"9661_weiwei_mabs_novo_chymo.raw",2777;2782,SGALTSGVH,9,99.9,99.9,9,414.71368,2,14.7044,6212751.5,827.4137,-1.1,"",100 100 100 100 100 100 100 100 100,MIX,SGALTSGVH,1268
"9661_weiwei_mabs_novo_chymo.raw",6831;6836,KTTPPVLDSDGSF,13,99.9,99.9,13,682.33929,2,36.6478,2.2263128E7,1362.6666,-1.9,"",99 100 100 100 100 100 100 100 100 100 100 100 100,MIX,KTTPPVLDSDGSF,5953
"9661_weiwei_mabs_novo_chymo.raw",2628;2631,YC(+58.01)QHSY,6,99.9,99.9,6,429.65768,2,12.8731,1.723695E7,857.3014,-0.7,"Carboxymethyl",100 100 100 100 100 100,MIX,YC(+58.01)QHSY,1569
"9661_weiwei_mabs_novo_chymo.raw",7151;7157,VSNKALPAPLEK,12,99.9,99.9,12,422.91812,3,37.0323,3034367.8,1265.7343,-1.4,"",100 100 100 100 100 100 100 100 100 100 100 100,MIX,VSNKALPAPLEK,1414
"9661_weiwei_mabs_novo_chymo.raw",8385;8391,SC(+58.01)SVM(+15.99)HEALHNHY,13,99.9,99.9,13,401.16647,4,44.6015,7.0465096E7,1600.6399,-1.9,"Carboxymethyl; Oxidation (M)",100 100 100 100 100 100 100 100 100 100 100 100 100,MIX,SC(+58.01)SVM(+15.99)HEALHNHY,1043
"9661_weiwei_mabs_novo_chymo.raw",10970,LAHLFSDDEKFY,12,99.9,99.9,12,495.57254,3,56.1511,3.4052484E7,1483.6984,-1.7,"",100 100 100 100 100 100 100 100 100 100 100 100,ETHCD,LAHLFSDDEKFY,2942
"9661_weiwei_mabs_novo_chymo.raw",4367;4371,SSPVTKSF,8,99.9,99.9,8,426.72595,2,22.5410,,851.4388,-1.8,"",100 100 100 100 100 100 100 100,MIX,SSPVTKSF,0
"9661_weiwei_mabs_novo_chymo.raw",10971,LAHLFSDDEKFY,12,99.9,99.9,12,495.57254,3,56.1511,3.4052484E7,1483.6984,-1.7,"",100 100 100 100 100 100 100 100 100 100 100 100,ETHCD,LAHLFSDDEKFY,2942
"9661_weiwei_mabs_novo_chymo.raw",10482;10490,LSAHNGDTNYAEKVQDRLTL,20,99.9,99.9,20,562.03442,4,53.8517,,2244.1135,-2.2,"",100 100 100 100 100 100 100 100 100 100 100 100 100 100 100 100 99 100 100 100,MIX,LSAHNGDTNYAEKVQDRLTL,0
"9661_weiwei_mabs_novo_chymo.raw",10972,LAHLFSDDEKFY,12,99.9,99.9,12,495.57254,3,56.1511,3.4052484E7,1483.6984,-1.7,"",100 100 100 100 100 100 100 100 100 100 100 100,ETHCD,LAHLFSDDEKFY,2942
"9661_weiwei_mabs_novo_chymo.raw",6579;6584,AC(+58.01)EVTHQGLSSPVTK,15,99.9,99.9,15,538.93024,3,35.3031,2.7658062E7,1613.7719,-1.9,"Carboxymethyl",100 100 100 100 100 100 100 100 100 100 100 99 100 100 100,MIX,AC(+58.01)EVTHQGLSSPVTK,3776
"9661_weiwei_mabs_novo_chymo.raw",3222;3229,SGALTSGVH,9,99.9,99.9,9,414.71368,2,14.7044,6212751.5,827.4137,-1.1,"",100 100 100 100 100 100 100 100 100,MIX,SGALTSGVH,1268
"9661_weiwei_mabs_novo_chymo.raw",4947;4951,TFRNY,5,99.9,99.9,5,350.67383,2,25.7710,,699.3340,-1.3,"",100 100 100 100 100,MIX,TFRNY,0"#;

const DATA_X_PLUS_SEP: &str = r"sep=	
Fraction	Source File	Feature	Peptide	Scan	Tag Length	Denovo Score	ALC (%)	length	m/z	z	RT	Predict RT	Area	Mass	ppm	PTM	local confidence (%)	tag (>=0%)	mode
1	20221121_F1_UM4_Peng0013_SA_EXT00_MIX3_SP3_LysN_1h.raw	6857	KVYAC(+58.01)EVTHQGLSSPVT	10653	17	99	99	17	938.9585	2	60.14	-	1.52E+09	1875.9036	-0.6	Carboxymethyl	96 100 100 100 100 100 100 100 100 99 100 100 100 100 100 99 99	KVYAC(+58.01)EVTHQGLSSPVT	HCD
1	20221121_F1_UM4_Peng0013_SA_EXT00_MIX3_SP3_LysN_1h.raw	6520	KVYAAEVTHQGLSSPVT	10831	17	99	99	17	893.9703	2	61.3	-	1.90E+07	1785.926	0		96 100 100 100 100 100 100 100 99 99 99 100 100 100 100 99 99	KVYAAEVTHQGLSSPVT	HCD
1	20221121_F1_UM4_Peng0013_SA_EXT00_MIX3_SP3_LysN_1h.raw	1606	KALPAPLEKTLSKA	13189	14	98	98	14	489.6371	3	74.21	-	2.78E+08	1465.8867	2		92 100 100 100 100 100 100 100 100 99 99 100 98 99	KALPAPLEKTLSKA	HCD
1	20221121_F1_UM4_Peng0013_SA_EXT00_MIX3_SP3_LysN_1h.raw	4843	KVDNALQSGNSQESVTEQDS	3940	20	98	98	20	712.6616	3	21.43	-	1.41E+08	2134.9614	0.8		88 100 100 100 100 100 99 100 100 99 98 98 100 100 100 100 100 99 100 99	KVDNALQSGNSQESVTEQDS	ETHCD
1	20221121_F1_UM4_Peng0013_SA_EXT00_MIX3_SP3_LysN_1h.raw	-	KALPAPLE	7853 7855	8	98	98	8	419.7561	2	43.85	-	0	837.496	2.1		90 100 100 100 100 100 100 100	KALPAPLE	HCD/ETHCD
1	20221121_F1_UM4_Peng0013_SA_EXT00_MIX3_SP3_LysN_1h.raw	-	KALPAPLE	8992 8994	8	98	98	8	419.756	2	50.47	-	0	837.496	1.7		89 100 100 100 100 100 100 100	KALPAPLE	HCD/ETHCD
1	20221121_F1_UM4_Peng0013_SA_EXT00_MIX3_SP3_LysN_1h.raw	-	KVDNALQSGNSQESVTEQDS	2890	20	98	98	20	712.6621	3	15.22	-	0	2134.9614	1.4		79 99 99 100 100 100 100 100 100 100 98 98 100 100 100 100 100 100 100 99	KVDNALQSGNSQESVTEQDS	ETHCD
1	20221121_F1_UM4_Peng0013_SA_EXT00_MIX3_SP3_LysN_1h.raw	5288	KADSSPVKAGVETTTPSKQSNN	10089	22	98	98	22	749.3798	3	56.94	-	9.32E+08	2245.1184	-0.4		91 99 99 100 100 97 96 98 99 99 99 100 100 100 100 99 98 98 97 99 99 100	KADSSPVKAGVETTTPSKQSNN	HCD
1	20221121_F1_UM4_Peng0013_SA_EXT00_MIX3_SP3_LysN_1h.raw	1830	KPSNTKVDK	10146	9	98	98	9	508.7907	2	57.19	-	1.13E+07	1015.5662	0.7		93 100 100 98 98 99 99 100 100	KPSNTKVDK	HCD
1	20221121_F1_UM4_Peng0013_SA_EXT00_MIX3_SP3_LysN_1h.raw	6566	KVYASEVTHQGLSSPVT	10495	17	98	98	17	901.9684	2	59.02	-	3.71E+07	1801.9209	0.7		93 100 100 100 100 100 100 100 99 98 90 97 100 100 100 99 99	KVYASEVTHQGLSSPVT	HCD
1	20221121_F1_UM4_Peng0013_SA_EXT00_MIX3_SP3_LysN_1h.raw	-	KALPAPLE	9507 9508	8	98	98	8	419.7558	2	53.55	-	0	837.496	1.4		87 100 100 100 100 100 100 100	KALPAPLE	HCD/ETHCD
1	20221121_F1_UM4_Peng0013_SA_EXT00_MIX3_SP3_LysN_1h.raw	-	KVYAC(+58.01)EVTHQGLSSPVT	10939	17	98	98	17	626.3096	3	61.85	-	0	1875.9036	1.8	Carboxymethyl	85 100 100 100 100 100 100 99 98 97 100 100 100 99 99 98 97	KVYAC(+58.01)EVTHQGLSSPVT	ETHCD
1	20221121_F1_UM4_Peng0013_SA_EXT00_MIX3_SP3_LysN_1h.raw	2378	KKVEPKSC(+58.01)D	9903	9	98	98	9	546.2742	2	55.87	-	1.55E+08	1090.5327	1	Carboxymethyl	90 100 100 100 99 99 100 98 99	KKVEPKSC(+58.01)D	HCD
1	20221121_F1_UM4_Peng0013_SA_EXT00_MIX3_SP3_LysN_1h.raw	2583	KFNWYVDGVEVHNA	13525	14	98	98	14	559.9401	3	76.19	-	6.99E+08	1676.7947	2.2		89 100 99 100 100 100 100 100 99 98 99 99 99 95	KFNWYVDGVEVHNA	ETHCD
1	20221121_F1_UM4_Peng0013_SA_EXT00_MIX3_SP3_LysN_1h.raw	-	KALPAPLE	8907 8908	8	98	98	8	419.7558	2	49.95	-	0	837.496	1.4		86 100 100 100 100 100 100 100	KALPAPLE	HCD/ETHCD
1	20221121_F1_UM4_Peng0013_SA_EXT00_MIX3_SP3_LysN_1h.raw	908	KVEPKSAD	2365	8	98	98	8	437.2383	2	12.02	-	1.78E+06	872.4603	2.1		91 100 100 99 99 99 99 100	KVEPKSAD	HCD
1	20221121_F1_UM4_Peng0013_SA_EXT00_MIX3_SP3_LysN_1h.raw	-	KFNWYVDGVEVHNA	13330	14	98	98	14	559.9392	3	75.22	-	0	1676.7947	0.7		83 100 100 100 100 100 100 100 100 100 100 100 99 95	KFNWYVDGVEVHNA	ETHCD
1	20221121_F1_UM4_Peng0013_SA_EXT00_MIX3_SP3_LysN_1h.raw	-	KADSSPVKAGVETTTPS	8144	17	98	98	17	558.9572	3	45.59	-	0	1673.8472	1.6		82 100 100 100 100 100 100 99 97 99 99 100 100 100 99 98 98	KADSSPVKAGVETTTPS	ETHCD
1	20221121_F1_UM4_Peng0013_SA_EXT00_MIX3_SP3_LysN_1h.raw	-	KADSSPVKAGVETTTPS	8084	17	98	98	17	558.9574	3	45.23	-	0	1673.8472	1.9		85 100 100 100 99 99 100 99 97 99 99 100 100 100 98 98 97	KADSSPVKAGVETTTPS	ETHCD";

const DATA_DB_PEPTIDE: &str = r"Peptide,-10lgP,Mass,Length,ppm,m/z,RT,Area Tryp-CID,Area Tryp EAD,Fraction,Scan,Source File,#Feature,#Feature Tryp-CID,#Feature Tryp EAD,Accession,PTM,AScore,Found By
ASQDVNTAVAWYQQKPGKAPK,133.29,2286.1755,21,2.4,572.5526,24.82,2.3211E2,2.4834E2,1,F1:150245,230629 1-Trypsin CID 06.mzML,3,2,1,Herceptin,,,PEAKS DB
VDNALQSGNSQESVTEQDSK,120.05,2134.9614,20,1.0,1068.4890,21.61,2.3684E3,5.8385E3,1,F1:141739,230629 1-Trypsin CID 06.mzML,4,3,1,Herceptin,,,PEAKS DB
HKVYAC(+57.02)EVTHQGLSSPVTK,111.34,2140.0735,19,1.9,536.0267,14.21,1.444E2,1.7678E2,1,F1:120669,230629 1-Trypsin CID 06.mzML,3,2,1,Herceptin,Carbamidomethylation,C6:Carbamidomethylation:1000.00,PEAKS DB
WQQGNVFSC(+57.02)SVMHEALHNHYTQK,100.40,2800.2598,23,1.8,701.0735,27.85,4.7485E1,6.7791E1,1,F1:157767,230629 1-Trypsin CID 06.mzML,4,3,1,Herceptin,Carbamidomethylation,C9:Carbamidomethylation:1000.00,PEAKS DB
VYAC(+57.02)EVTHQGLSSPVTK,99.81,1874.9197,17,0.0,938.4671,24.05,4.1398E2,1.0896E3,1,F1:148000,230629 1-Trypsin CID 06.mzML,3,2,1,Herceptin,Carbamidomethylation,C4:Carbamidomethylation:1000.00,PEAKS DB
WGGDGFYAMDYWGQGTLVTVSSASTK,98.11,2783.2537,26,-1.7,928.7570,36.88,3.0651E0,,1,F1:181155,230629 1-Trypsin CID 06.mzML,1,1,0,Herceptin,,,PEAKS DB
DIQM(+15.99)TQSPSSLSASVGDR,97.71,1893.8738,18,-0.3,947.9438,27.29,4.2605E2,8.659E2,1,F1:156563,230629 1-Trypsin CID 06.mzML,7,4,3,Herceptin,Oxidation (M),M4:Oxidation (M):1000.00,PEAKS DB
SGTASVVC(+57.02)LLNNFYPR,97.60,1796.8879,16,0.3,599.9701,36.54,1.2015E1,4.5262E1,1,F1:180470,230629 1-Trypsin CID 06.mzML,3,2,1,Herceptin,Carbamidomethylation,C8:Carbamidomethylation:1000.00,PEAKS DB
EVQLVESGGGLVQPGGSLR,96.67,1880.9955,19,-0.7,941.5044,29.26,4.3056E3,1.4147E4,1,F1:161445,230629 1-Trypsin CID 06.mzML,10,2,8,Herceptin,,,PEAKS DB
VDNALQSGNSQESVTEQDSKDSTYSLSSTLTLSK,95.79,3618.7019,34,1.7,1207.2433,28.97,1.4966E2,,1,F1:160475,230629 1-Trypsin CID 06.mzML,3,3,0,Herceptin,,,PEAKS DB
ASQDVNTAVAWYQQKPGK,94.88,1989.9908,18,2.7,498.5063,26.07,2.6249E3,8.616E3,1,F1:153458,230629 1-Trypsin CID 06.mzML,5,3,2,Herceptin,,,PEAKS DB
DIQMTQSPSSLSASVGDR,94.57,1877.8789,18,0.3,939.9470,27.29,1.2159E3,5.1458E3,1,F1:156481,230629 1-Trypsin CID 06.mzML,8,4,4,Herceptin,,,PEAKS DB
GFYPSDIAVEWESNGQPENNYK,94.55,2543.1240,22,1.1,848.7162,33.01,2.3514E2,3.8553E3,1,F1:170996,230629 1-Trypsin CID 06.mzML,3,2,1,Herceptin,,,PEAKS DB
ASQDVNTAVAW(+15.99)YQQKPGK,90.25,2005.9857,18,1.0,669.6699,23.55,3.9445E1,1.105E2,1,F1:146706,230629 1-Trypsin CID 06.mzML,12,7,5,Herceptin,Oxidation (HW),W11:Oxidation (HW):1000.00,PEAKS DB
TTPPVLDSDGSFFLYSK,89.61,1872.9146,17,-0.2,937.4644,33.77,2.6393E3,2.0245E4,1,F1:172987,230629 1-Trypsin CID 06.mzML,4,2,2,Herceptin,,,PEAKS DB
FNW(+15.99)YVDGVEVHNAK,88.93,1692.7896,14,0.3,847.4023,29.08,5.181E2,1.1824E3,1,F1:160805,230629 1-Trypsin CID 06.mzML,18,13,5,Herceptin,Oxidation (HW),W3:Oxidation (HW):113.97,PEAKS DB
VVSVLTVLHQDWLNGK,88.68,1806.9991,16,-0.1,904.5068,35.17,2.394E2,1.0949E3,1,F1:176568,230629 1-Trypsin CID 06.mzML,3,2,1,Herceptin,,,PEAKS DB
DSTYSLSSTLTLSK,88.53,1501.7511,14,-0.4,751.8825,30.22,1.0615E2,1.1102E3,1,F1:163866,230629 1-Trypsin CID 06.mzML,6,4,2,Herceptin,,,PEAKS DB
FNWYVDGVEVHNAK,88.47,1676.7947,14,0.5,839.4050,29.08,5.6654E3,1.6995E4,1,F1:160881,230629 1-Trypsin CID 06.mzML,6,3,3,Herceptin,,,PEAKS DB";

const DATA_DB_PSM: &str = r"Peptide,-10lgP,Mass,Length,ppm,m/z,Z,RT,Area,Fraction,Id,Scan,from Chimera,Source File,Accession,PTM,AScore,Found By
ASQDVNTAVAWYQQKPGKAPK,133.29,2286.1755,21,2.4,572.5526,4,24.82,2.0118E2,1,1551,F1:150245,No,230629 1-Trypsin CID 06.mzML,Herceptin,,,PEAKS DB
ASQDVNTAVAWYQQKPGKAPK,100.35,2286.1755,21,-0.5,763.0654,3,24.82,3.0934E1,1,1552,F1:150301,No,230629 1-Trypsin CID 06.mzML,Herceptin,,,PEAKS DB
ASQDVNTAVAWYQQKPGKAPK,30.99,2286.1755,21,1.4,572.5519,4,24.99,2.4834E2,2,13517,F2:87375,No,230629 1-Trypsin EAD 02.mzML,Herceptin,,,PEAKS DB
VDNALQSGNSQESVTEQDSK,120.05,2134.9614,20,1.0,1068.4890,2,21.61,5.4455E2,1,746,F1:141739,No,230629 1-Trypsin CID 06.mzML,Herceptin,,,PEAKS DB
VDNALQSGNSQESVTEQDSK,108.86,2134.9614,20,1.5,712.6622,3,21.61,1.823E3,1,745,F1:141693,No,230629 1-Trypsin CID 06.mzML,Herceptin,,,PEAKS DB
VDNALQSGNSQESVTEQDSK,108.49,2134.9614,20,0.2,712.6613,3,21.83,0,1,7135,F1:142330,No,230629 1-Trypsin CID 06.mzML,Herceptin,,,PEAKS DB
VDNALQSGNSQESVTEQDSK,95.26,2134.9614,20,2.0,712.6625,3,21.37,0,1,7045,F1:141088,No,230629 1-Trypsin CID 06.mzML,Herceptin,,,PEAKS DB
VDNALQSGNSQESVTEQDSK,91.13,2134.9614,20,-0.4,712.6608,3,21.40,0,1,7046,F1:141182,No,230629 1-Trypsin CID 06.mzML,Herceptin,,,PEAKS DB
VDNALQSGNSQESVTEQDSK,80.87,2134.9614,20,-0.1,712.6610,3,14.25,8.4422E-1,1,616,F1:120823,No,230629 1-Trypsin CID 06.mzML,Herceptin,,,PEAKS DB
VDNALQSGNSQESVTEQDSK,24.39,2134.9614,20,-0.5,712.6607,3,21.64,5.8385E3,2,12319,F2:81886,No,230629 1-Trypsin EAD 02.mzML,Herceptin,,,PEAKS DB
VDNALQSGNSQESVTEQDSK,19.22,2134.9614,20,0.2,712.6613,3,21.35,0,2,22568,F2:81360,No,230629 1-Trypsin EAD 02.mzML,Herceptin,,,PEAKS DB
VDNALQSGNSQESVTEQDSK,18.53,2134.9614,20,-0.7,534.7473,4,21.60,0,2,22646,F2:81777,No,230629 1-Trypsin EAD 02.mzML,Herceptin,,,PEAKS DB
HKVYAC(+57.02)EVTHQGLSSPVTK,111.34,2140.0735,19,1.9,536.0267,4,14.21,1.2924E2,1,602,F1:120669,No,230629 1-Trypsin CID 06.mzML,Herceptin,Carbamidomethylation,C6:Carbamidomethylation:1000.00,PEAKS DB
HKVYAC(+57.02)EVTHQGLSSPVTK,98.89,2140.0735,19,3.3,714.3674,3,14.21,1.5163E1,1,608,F1:120715,No,230629 1-Trypsin CID 06.mzML,Herceptin,Carbamidomethylation,C6:Carbamidomethylation:1000.00,PEAKS DB
HKVYAC(+57.02)EVTHQGLSSPVTK,19.78,2140.0735,19,0.2,536.0258,4,14.22,1.7678E2,2,12131,F2:69011,No,230629 1-Trypsin EAD 02.mzML,Herceptin,Carbamidomethylation,C6:Carbamidomethylation:1000.00,PEAKS DB
WQQGNVFSC(+57.02)SVMHEALHNHYTQK,100.40,2800.2598,23,1.8,701.0735,4,27.85,2.1062E1,1,2906,F1:157767,No,230629 1-Trypsin CID 06.mzML,Herceptin,Carbamidomethylation,C9:Carbamidomethylation:1000.00,PEAKS DB
WQQGNVFSC(+57.02)SVMHEALHNHYTQK,90.42,2800.2598,23,-1.4,561.0585,5,27.85,2.3978E1,1,2905,F1:157803,No,230629 1-Trypsin CID 06.mzML,Herceptin,Carbamidomethylation,C9:Carbamidomethylation:1000.00,PEAKS DB
WQQGNVFSC(+57.02)SVMHEALHNHYTQK,42.54,2800.2598,23,-1.5,934.4258,3,27.83,2.4453E0,1,2897,F1:157770,Yes,230629 1-Trypsin CID 06.mzML,Herceptin,Carbamidomethylation,C9:Carbamidomethylation:1000.00,PEAKS DB
WQQGNVFSC(+57.02)SVMHEALHNHYTQK,34.88,2800.2598,23,-12.0,701.0638,4,28.42,0,1,9493,F1:159270,No,230629 1-Trypsin CID 06.mzML,Herceptin,Carbamidomethylation,C9:Carbamidomethylation:1000.00,PEAKS DB";

const DATA_DB_PROTEIN_PEPTIDE: &str = r"Protein Group,Protein ID,Protein Accession,Peptide,Unique,-10lgP,Mass,Length,ppm,m/z,z,RT,Area Tryp-CID,Area Tryp EAD,Fraction,Scan,Source File,#Feature,#Feature Tryp-CID,#Feature Tryp EAD,Start,End,PTM,AScore,Found By
1,1,Herceptin,R.ASQDVNTAVAWYQQKPGKAPK.L,Y,133.29,2286.1755,21,2.4,572.5526,4,24.82,2.3211E2,2.4834E2,1,F1:150245,230629 1-Trypsin CID 06.mzML,3,2,1,475,495,,,PEAKS DB
1,1,Herceptin,K.VDNALQSGNSQESVTEQDSK.D,Y,120.05,2134.9614,20,1.0,1068.4890,2,21.61,2.3684E3,5.8385E3,1,F1:141739,230629 1-Trypsin CID 06.mzML,4,3,1,600,619,,,PEAKS DB
1,1,Herceptin,K.HKVYAC(+57.02)EVTHQGLSSPVTK.S,Y,111.34,2140.0735,19,1.9,536.0267,4,14.21,1.444E2,1.7678E2,1,F1:120669,230629 1-Trypsin CID 06.mzML,3,2,1,639,657,Carbamidomethylation,C6:Carbamidomethylation:1000.00,PEAKS DB
1,1,Herceptin,R.WQQGNVFSC(+57.02)SVMHEALHNHYTQK.S,Y,100.40,2800.2598,23,1.8,701.0735,4,27.85,4.7485E1,6.7791E1,1,F1:157767,230629 1-Trypsin CID 06.mzML,4,3,1,420,442,Carbamidomethylation,C9:Carbamidomethylation:1000.00,PEAKS DB
1,1,Herceptin,K.VYAC(+57.02)EVTHQGLSSPVTK.S,Y,99.81,1874.9197,17,0.0,938.4671,2,24.05,4.1398E2,1.0896E3,1,F1:148000,230629 1-Trypsin CID 06.mzML,3,2,1,641,657,Carbamidomethylation,C4:Carbamidomethylation:1000.00,PEAKS DB
1,1,Herceptin,R.WGGDGFYAMDYWGQGTLVTVSSASTK.G,Y,98.11,2783.2537,26,-1.7,928.7570,3,36.88,3.0651E0,,1,F1:181155,230629 1-Trypsin CID 06.mzML,1,1,0,99,124,,,PEAKS DB
1,1,Herceptin,K.DIQM(+15.99)TQSPSSLSASVGDR.V,Y,97.71,1893.8738,18,-0.3,947.9438,2,27.29,4.2605E2,8.659E2,1,F1:156563,230629 1-Trypsin CID 06.mzML,7,4,3,451,468,Oxidation (M),M4:Oxidation (M):1000.00,PEAKS DB
1,1,Herceptin,K.SGTASVVC(+57.02)LLNNFYPR.E,Y,97.60,1796.8879,16,0.3,599.9701,3,36.54,1.2015E1,4.5262E1,1,F1:180470,230629 1-Trypsin CID 06.mzML,3,2,1,577,592,Carbamidomethylation,C8:Carbamidomethylation:1000.00,PEAKS DB
1,1,Herceptin,EVQLVESGGGLVQPGGSLR.L,Y,96.67,1880.9955,19,-0.7,941.5044,2,29.26,4.3056E3,1.4147E4,1,F1:161445,230629 1-Trypsin CID 06.mzML,10,2,8,1,19,,,PEAKS DB
1,1,Herceptin,K.VDNALQSGNSQESVTEQDSKDSTYSLSSTLTLSK.A,Y,95.79,3618.7019,34,1.7,1207.2433,3,28.97,1.4966E2,,1,F1:160475,230629 1-Trypsin CID 06.mzML,3,3,0,600,633,,,PEAKS DB
1,1,Herceptin,R.ASQDVNTAVAWYQQKPGK.A,Y,94.88,1989.9908,18,2.7,498.5063,4,26.07,2.6249E3,8.616E3,1,F1:153458,230629 1-Trypsin CID 06.mzML,5,3,2,475,492,,,PEAKS DB
1,1,Herceptin,K.DIQMTQSPSSLSASVGDR.V,Y,94.57,1877.8789,18,0.3,939.9470,2,27.29,1.2159E3,5.1458E3,1,F1:156481,230629 1-Trypsin CID 06.mzML,8,4,4,451,468,,,PEAKS DB
1,1,Herceptin,K.GFYPSDIAVEWESNGQPENNYK.T,Y,94.55,2543.1240,22,1.1,848.7162,3,33.01,2.3514E2,3.8553E3,1,F1:170996,230629 1-Trypsin CID 06.mzML,3,2,1,374,395,,,PEAKS DB
1,1,Herceptin,R.ASQDVNTAVAW(+15.99)YQQKPGK.A,Y,90.25,2005.9857,18,1.0,669.6699,3,23.55,3.9445E1,1.105E2,1,F1:146706,230629 1-Trypsin CID 06.mzML,12,7,5,475,492,Oxidation (HW),W11:Oxidation (HW):1000.00,PEAKS DB
1,1,Herceptin,K.TTPPVLDSDGSFFLYSK.L,Y,89.61,1872.9146,17,-0.2,937.4644,2,33.77,2.6393E3,2.0245E4,1,F1:172987,230629 1-Trypsin CID 06.mzML,4,2,2,396,412,,,PEAKS DB
1,1,Herceptin,K.FNW(+15.99)YVDGVEVHNAK.T,Y,88.93,1692.7896,14,0.3,847.4023,2,29.08,5.181E2,1.1824E3,1,F1:160805,230629 1-Trypsin CID 06.mzML,18,13,5,278,291,Oxidation (HW),W3:Oxidation (HW):113.97,PEAKS DB
1,1,Herceptin,R.VVSVLTVLHQDWLNGK.E,Y,88.68,1806.9991,16,-0.1,904.5068,2,35.17,2.394E2,1.0949E3,1,F1:176568,230629 1-Trypsin CID 06.mzML,3,2,1,305,320,,,PEAKS DB
1,1,Herceptin,K.DSTYSLSSTLTLSK.A,Y,88.53,1501.7511,14,-0.4,751.8825,2,30.22,1.0615E2,1.1102E3,1,F1:163866,230629 1-Trypsin CID 06.mzML,6,4,2,620,633,,,PEAKS DB
1,1,Herceptin,K.FNWYVDGVEVHNAK.T,Y,88.47,1676.7947,14,0.5,839.4050,2,29.08,5.6654E3,1.6995E4,1,F1:160881,230629 1-Trypsin CID 06.mzML,6,3,3,278,291,,,PEAKS DB";
